<?php

namespace JetBackup\Alert;

use Exception;
use JetBackup\Ajax\Ajax;
use JetBackup\CLI\CLI;
use JetBackup\Data\Engine;
use JetBackup\Data\SleekStore;
use JetBackup\Entities\Util;
use JetBackup\JetBackup;
use SleekDB\Exceptions\InvalidArgumentException;
use SleekDB\Exceptions\IOException;
use SleekDB\QueryBuilder;

if (!defined( '__JETBACKUP__')) die('Direct access is not allowed');

class Alert extends Engine {

	const COLLECTION = 'alerts';
	
	const UNIQUE_ID = 'unique_id';
	const CREATED = 'created';
	const TITLE = 'title';
	const MESSAGE = 'message';
	const LEVEL = 'level';

	const LEVEL_INFORMATION = 1;
	const LEVEL_WARNING = 2;
	const LEVEL_CRITICAL = 4;
	
	const LEVEL_NAMES = [
		self::LEVEL_INFORMATION     => 'Information',
		self::LEVEL_WARNING         => 'Warning',
		self::LEVEL_CRITICAL        => 'Critical',
	];
	
	public function __construct($_id=null) {
		parent::__construct(self::COLLECTION);
		if($_id) $this->_loadById((int) $_id);
	}

	public function setUniqueId($id) { $this->set(self::UNIQUE_ID, $id); }
	public function getUniqueId():string { return $this->get(self::UNIQUE_ID); }

	public function setCreated($value):void { $this->set(self::CREATED, $value); }
	public function getCreated() { return $this->get(self::CREATED); }

	public function setTitle($value):void { $this->set(self::TITLE, $value); }
	public function getTitle() { return $this->get(self::TITLE); }

	public function setMessage($value):void { $this->set(self::MESSAGE, $value); }
	public function getMessage() { return $this->get(self::MESSAGE); }

	public function setLevel(int $value):void { $this->set(self::LEVEL, $value); }
	public function getLevel():int { return (int) $this->get(self::LEVEL); }

	public function save():void {
		if(!$this->getUniqueId()) $this->setUniqueId(Util::generateUniqueId());
		parent::save();
	}

	public static function db():SleekStore {
		return new SleekStore(self::COLLECTION);
	}

	public static function query():QueryBuilder {
		return self::db()->createQueryBuilder();
	}

	public static function getTotalAlerts():int {
		return count(self::query()
             ->select([JetBackup::ID_FIELD])
             ->getQuery()
             ->fetch());
	}

	public static function getTotalCriticalAlerts():int {
		return count(self::query()
		                 ->select([JetBackup::ID_FIELD])
						->where([self::LEVEL, '=', self::LEVEL_CRITICAL])
		                 ->getQuery()
		                 ->fetch());
	}

	/**
	 * @throws InvalidArgumentException
	 * @throws IOException
	 */
	public static function clearAlerts():void {
		self::query()->getQuery()->delete();
	}
	
	public static function add(string $title, string $message, int $level):void {
		$alert = new Alert();
		$alert->setCreated(time());
		$alert->setTitle($title);
		$alert->setMessage($message);
		$alert->setLevel($level);
		$alert->setEngine(Engine::ENGINE_WP);
		$alert->save();
	}

	/**
	 * @throws Exception
	 */
	public function getDisplay(): array {
		return [
			JetBackup::ID_FIELD => $this->getId(),
			self::UNIQUE_ID     => $this->getUniqueId(),
			self::CREATED       => $this->getCreated(),
			self::TITLE         => $this->getTitle(),
			self::MESSAGE       => $this->getMessage(),
			Engine::ENGINE      => $this->getEngine(),
			'engine_name'       => $this->getEngineName(),
			self::LEVEL         => $this->getLevel(),
		];
	}

	/**
	 * @throws Exception
	 */
	public function getDisplayCLI(): array {
		return [
			'ID'        => $this->getId(),
			'Created'   => CLI::date($this->getCreated()),
			'Title'     => $this->getTitle(),
			'Message'   => $this->getMessage(),
			'Engine'    => $this->getEngineName(),
			'Level'     => self::LEVEL_NAMES[$this->getLevel()],
		];
	}
}